"""Main stb-tester python module. Intended to be used with `stbt run`.

See `man stbt` and http://stb-tester.com for documentation.

Copyright 2012-2013 YouView TV Ltd and contributors.
Copyright 2013-2018 stb-tester.com Ltd.
License: LGPL v2.1 or (at your option) any later version (see
https://github.com/stb-tester/stb-tester/blob/master/LICENSE for details).
"""

from __future__ import annotations

import typing
from contextlib import contextmanager
from typing import ContextManager, Iterator, Optional

from _stbt import android
from _stbt.black import (
    is_screen_black)
from _stbt.config import (
    ConfigurationError,
    get_config)
from _stbt.diff import (
    BGRDiff,
    Differ,
    GrayscaleDiff,
    MotionResult)
from _stbt.frameobject import (
    for_object_repository,
    FrameObject)
from _stbt.grid import (
    Grid)
from _stbt.imgutils import (
    Color,
    crop,
    find_file,
    Frame,
    Image,
    load_image,
    save_frame)
from _stbt.keyboard import (
    Keyboard)
from _stbt.logging import (
    debug)
from _stbt.mask import (
    load_mask,
    Mask,
    MaskTypes)
from _stbt.match import (
    ConfirmMethod,
    match,
    match_all,
    MatchMethod,
    MatchParameters,
    MatchResult,
    MatchTimeout,
    wait_for_match)
from _stbt.motion import (
    detect_motion,
    MotionTimeout,
    wait_for_motion)
from _stbt.multipress import (
    MultiPress)
from _stbt.ocr import (
    apply_ocr_corrections,
    match_text,
    ocr,
    ocr_eq,
    OcrEngine,
    OcrMode,
    set_global_ocr_corrections,
    TextMatchResult)
from _stbt.precondition import (
    as_precondition,
    PreconditionError)
from _stbt.transition import (
    press_and_wait,
    Transition,
    TransitionStatus,
    wait_for_transition_to_end)
from _stbt.types import (
    Direction,
    Keypress,
    NoVideo,
    PDU,
    Position,
    Region,
    Size,
    UITestError,
    UITestFailure)
from _stbt.wait import (
    wait_until)

__all__ = [
    "android",
    "apply_ocr_corrections",
    "as_precondition",
    "BGRDiff",
    "Color",
    "ConfigurationError",
    "ConfirmMethod",
    "crop",
    "debug",
    "detect_motion",
    "Differ",
    "Direction",
    "draw_text",
    "find_file",
    "for_object_repository",
    "Frame",
    "FrameObject",
    "frames",
    "get_config",
    "get_frame",
    "GrayscaleDiff",
    "Grid",
    "Image",
    "is_screen_black",
    "Keyboard",
    "Keypress",
    "last_keypress",
    "load_image",
    "load_mask",
    "Mask",
    "MaskTypes",
    "match",
    "match_all",
    "match_text",
    "MatchMethod",
    "MatchParameters",
    "MatchResult",
    "MatchTimeout",
    "MotionResult",
    "MotionTimeout",
    "MultiPress",
    "NoVideo",
    "ocr",
    "ocr_eq",
    "OcrEngine",
    "OcrMode",
    "PDU",
    "Position",
    "PreconditionError",
    "press",
    "press_and_wait",
    "press_until_match",
    "pressing",
    "Region",
    "save_frame",
    "set_global_ocr_corrections",
    "Size",
    "TextMatchResult",
    "Transition",
    "TransitionStatus",
    "UITestError",
    "UITestFailure",
    "wait_for_match",
    "wait_for_motion",
    "wait_for_transition_to_end",
    "wait_until",
]


from _stbt.imgutils import ImageT
from _stbt.types import KeyT, RegionT
if typing.TYPE_CHECKING:
    import _stbt.core


TEST_PACK_ROOT: "str|None" = None


# Functions available to stbt scripts
# ===========================================================================


def last_keypress() -> "Keypress|None":
    """Returns information about the last key-press sent to the device under
    test.

    See the return type of `stbt.press`.

    Added in v32.
    """
    return _dut.last_keypress()


def press(
    key: KeyT, interpress_delay_secs: Optional[float] = None,
    hold_secs: Optional[float] = None
) -> Keypress:
    """Send the specified key-press to the device under test.

    :param str key:
        The name of the key/button.

        If you are using infrared control, this is a key name from your
        lircd.conf configuration file.

        If you are using HDMI CEC control, see the available key names
        `here <https://stb-tester.com/kb/cec-names>`__. Note that some devices
        might not understand every CEC command in that list.

    :type interpress_delay_secs: int or float
    :param interpress_delay_secs:
        The minimum time to wait after a previous key-press, in order to
        accommodate the responsiveness of the device-under-test.

        This defaults to 0.3. You can override the global default value by
        setting ``interpress_delay_secs`` in the ``[press]`` section of
        :ref:`.stbt.conf`.

    :type hold_secs: int or float
    :param hold_secs:
        Hold the key down for the specified duration (in seconds). Currently
        this is implemented for the infrared, HDMI CEC, and Roku controls.
        There is a maximum limit of 60 seconds.

    :returns:
        A `stbt.Keypress` object with information about the keypress that was
        sent.

    * Changed in v33: The ``key`` argument can be an Enum (we'll use the Enum's
      value, which must be a string).
    """
    return _dut.press(key, interpress_delay_secs, hold_secs)


def pressing(
    key: KeyT, interpress_delay_secs: Optional[float] = None
) -> ContextManager[Keypress]:
    """Context manager that will press and hold the specified key for the
    duration of the ``with`` code block.

    For example, this will hold KEY_RIGHT until ``wait_for_match`` finds a
    match or times out::

        with stbt.pressing("KEY_RIGHT"):
            stbt.wait_for_match("last-page.png")

    The same limitations apply as `stbt.press`'s ``hold_secs`` parameter.
    """
    return _dut.pressing(key, interpress_delay_secs)


def draw_text(text: str, duration_secs: float = 3) -> None:
    """Write the specified text to the output video.

    :param str text: The text to write.

    :param duration_secs: The number of seconds to display the text.
    :type duration_secs: int or float
    """
    debug(text)
    return _dut.draw_text(text, duration_secs)


def press_until_match(
    key: KeyT,
    image: ImageT,
    interval_secs: Optional[float] = None,
    max_presses: Optional[int] = None,
    match_parameters: Optional[MatchParameters] = None,
    region: RegionT = Region.ALL,
) -> MatchResult:
    """Call `press` as many times as necessary to find the specified image.

    :param key: See `press`.

    :param image: See `match`.

    :type interval_secs: int or float
    :param interval_secs:
        The number of seconds to wait for a match before pressing again.
        Defaults to 3.

        You can override the global default value by setting ``interval_secs``
        in the ``[press_until_match]`` section of :ref:`.stbt.conf`.

    :param int max_presses:
        The number of times to try pressing the key and looking for the image
        before giving up and raising `MatchTimeout`. Defaults to 10.

        You can override the global default value by setting ``max_presses``
        in the ``[press_until_match]`` section of :ref:`.stbt.conf`.

    :param match_parameters: See `match`.
    :param region: See `match`.

    :returns: `MatchResult` when the image is found.
    :raises: `MatchTimeout` if no match is found after ``timeout_secs`` seconds.
    """
    return _dut.press_until_match(
        key, image, interval_secs, max_presses, match_parameters, region)


def frames(timeout_secs: Optional[float] = None) -> Iterator[Frame]:
    """Generator that yields video frames captured from the device-under-test.

    For example::

        for frame in stbt.frames():
            # Do something with each frame here.
            # Remember to add a termination condition to `break` or `return`
            # from the loop, or specify `timeout_secs` — otherwise you'll have
            # an infinite loop!
            ...

    See also `stbt.get_frame`.

    :type timeout_secs: int or float or None
    :param timeout_secs:
      A timeout in seconds. After this timeout the iterator will be exhausted.
      That is, a ``for`` loop like ``for f in stbt.frames(timeout_secs=10)``
      will terminate after 10 seconds. If ``timeout_secs`` is ``None`` (the
      default) then the iterator will yield frames forever but you can stop
      iterating (for example with ``break``) at any time.

    :rtype: Iterator[stbt.Frame]
    :returns:
      An iterator of frames in OpenCV format (`stbt.Frame`).
    """
    return _dut.frames(timeout_secs)


def get_frame() -> Frame:
    """Grabs a video frame from the device-under-test.

    :rtype: stbt.Frame
    :returns: The most recent video frame in OpenCV format.

    Most Stb-tester APIs (`stbt.match`, `stbt.FrameObject` constructors, etc.)
    will call ``get_frame`` if a frame isn't specified explicitly.

    If you call ``get_frame`` twice very quickly (faster than the video-capture
    framerate) you might get the same frame twice. To block until the next
    frame is available, use `stbt.frames`.

    To save a frame to disk pass it to :ocv:pyfunc:`cv2.imwrite`. Note that any
    file you write to the current working directory will appear as an artifact
    in the test-run results.
    """
    return _dut.get_frame()


# Internal
# ===========================================================================

class UnconfiguredDeviceUnderTest():
    def last_keypress(self):
        return None

    def press(self, *args, **kwargs):
        raise RuntimeError(
            "stbt.press isn't configured to run on your hardware")

    def pressing(self, *args, **kwargs):
        raise RuntimeError(
            "stbt.pressing isn't configured to run on your hardware")

    def draw_text(self, *args, **kwargs):
        # Unlike the others this has no external side-effects on the device
        # under test. `stbt.draw_text` already logs it before calling
        # `_dut.draw_text`. Really this shouldn't belong in the DUT at all.
        pass

    def press_until_match(self, *args, **kwargs):
        raise RuntimeError(
            "stbt.press_until_match isn't configured to run on your hardware")

    def frames(self, *args, **kwargs):
        raise RuntimeError(
            "stbt.frames isn't configured to run on your hardware")

    def get_frame(self, *args, **kwargs):
        raise RuntimeError(
            "stbt.get_frame isn't configured to run on your hardware")


_dut: "_stbt.core.DeviceUnderTest | UnconfiguredDeviceUnderTest" = (
    UnconfiguredDeviceUnderTest())


@contextmanager
def _set_dut_singleton(dut):
    global _dut
    old_dut = dut
    try:
        _dut = dut
        yield dut
    finally:
        _dut = old_dut



"""Copyright 2014-2019 Stb-tester.com Ltd.

This file shouldn't depend on anything else in stbt.
"""

import errno
import os
import re
import tempfile
from contextlib import contextmanager
from shutil import rmtree


def mkdir_p(d):
    """Python 3.2 has an optional argument to os.makedirs called exist_ok.  To
    support older versions of python we can't use this and need to catch
    exceptions"""
    try:
        os.makedirs(d)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(d) \
                and os.access(d, os.R_OK | os.W_OK):
            return
        else:
            raise


def rm_f(filename):
    """Like ``rm -f``, it ignores errors if the file doesn't exist."""
    try:
        os.remove(filename)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise


@contextmanager
def named_temporary_directory(
        suffix='', prefix='tmp', dir=None):  # pylint:disable=redefined-builtin
    dirname = tempfile.mkdtemp(suffix, prefix, dir)
    try:
        yield dirname
    finally:
        rmtree(dirname)


@contextmanager
def scoped_curdir():
    with named_temporary_directory() as tmpdir:
        olddir = os.path.abspath(os.curdir)
        os.chdir(tmpdir)
        try:
            yield olddir
        finally:
            os.chdir(olddir)


@contextmanager
def scoped_process(process):
    try:
        yield process
    finally:
        if process.poll() is None:
            process.kill()
            process.wait()


def find_import_name(filename):
    """
    To import an arbitrary filename we need to set PYTHONPATH and we need to
    know the name of the module we're importing.  This is complicated by Python
    packages: for a directory structure like this:

        tests/package/a.py
        tests/package/__init__.py

    we want to add `tests` to `PYTHONPATH` (`sys.path`) and `import package.a`.
    This function traverses the directories to work out what `PYTHONPATH` and
    the module name should be returning them as a tuple.
    """
    import_dir, module_file = os.path.split(os.path.abspath(filename))
    import_name, module_ext = os.path.splitext(module_file)
    if module_ext != '.py':
        raise ImportError("Invalid module filename '%s'" % filename)

    while os.path.exists(os.path.join(import_dir, "__init__.py")):
        import_dir, s = os.path.split(import_dir)
        import_name = "%s.%s" % (s, import_name)
    return import_dir, import_name


_component_re = re.compile(r'(\d+ | [a-z]+ | \.)', re.VERBOSE)


def LooseVersion(vstring):
    """Copied from `distutils.version.LooseVersion`.

    Note that (like distutils.version.LooseVersion) this simply sorts
    lexicographically according to the "." or "-" separated components in the
    version string:

    >>> LooseVersion("4.0.0-beta.1")
    [4, 0, 0, '-', 'beta', 1]
    >>> (LooseVersion('4.0.0-beta.1') > LooseVersion('4.0.0'))
    True
    """

    components = [x for x in _component_re.split(vstring)
                  if x and x != '.']
    for i, obj in enumerate(components):
        try:
            components[i] = int(obj)
        except ValueError:
            pass
    return components


def to_bytes(text):
    if isinstance(text, str):
        return text.encode("utf-8", errors="backslashreplace")
    elif isinstance(text, bytes):
        return text
    else:
        raise TypeError("Unexpected type %s" % type(text))


def to_unicode(text):
    if isinstance(text, bytes):
        return text.decode("utf-8", errors="replace")
    else:
        return str(text)



"""Copyright 2020 Stb-tester.com Ltd."""

from __future__ import annotations

import re
import time
from typing import Optional

from .types import KeyT


class MultiPress():
    """Helper for entering text using `multi-press`_ on a numeric keypad.

    In some apps, the search page allows entering text by pressing the keys on
    the remote control's numeric keypad: press the number "2" once for "A",
    twice for "B", etc.::

        1.,     ABC2    DEF3
        GHI4    JKL5    MNO6
        PQRS7   TUV8    WXYZ9
              [space]0

    To enter text with this mechanism, create an instance of this class and
    call its ``enter_text`` method. For example::

        multipress = stbt.MultiPress()
        multipress.enter_text("teletubbies")

    The constructor takes the following parameters:

    :param dict key_mapping:
        The mapping from number keys to letters. The default mapping is::

            {
                "KEY_0": " 0",
                "KEY_1": "1.,",
                "KEY_2": "abc2",
                "KEY_3": "def3",
                "KEY_4": "ghi4",
                "KEY_5": "jkl5",
                "KEY_6": "mno6",
                "KEY_7": "pqrs7",
                "KEY_8": "tuv8",
                "KEY_9": "wxyz9",
            }

        This matches the arrangement of digits A-Z from ITU E.161 / ISO 9995-8.

        The value you pass in this parameter is merged with the default mapping.
        For example to override the punctuation characters you can specify
        ``key_mapping={"KEY_1": "@1.,-_"}``.

        The dict's key names must match the remote-control key names accepted
        by `stbt.press`. The dict's values are a string or sequence of the
        corresponding letters, in the order that they are entered when pressing
        that key.

    :param float interpress_delay_secs:
        The time to wait between every key-press, in seconds. This defaults to
        0.3, the same default as `stbt.press`.

    :param float interletter_delay_secs:
        The time to wait between letters on the same key, in seconds. For
        example, to enter "AB" you need to press key "2" once, then wait, then
        press it again twice. If you don't wait, the device-under-test would
        see three consecutive keypresses which mean the letter "C".

    .. _multi-press: https://en.wikipedia.org/wiki/Multi-tap
    """

    def __init__(
        self,
        key_mapping: Optional[dict[KeyT, str]] = None,
        interpress_delay_secs: Optional[float] = None,
        interletter_delay_secs: float = 1,
    ):

        mapping = _parse_mapping_from_docstring(MultiPress.__doc__)
        if key_mapping is not None:
            mapping.update(key_mapping)
        self.keys = _letters_to_keys(mapping)

        self.interpress_delay_secs = interpress_delay_secs
        self.interletter_delay_secs = interletter_delay_secs

    def enter_text(self, text: str) -> None:
        """Enter the specified text using multi-press on the numeric keypad.

        :param str text:
            The text to enter. The case doesn't matter (uppercase and lowercase
            are treated the same).
        """

        from stbt_core import debug, press

        text = text.lower()

        # Raise exception early, so we don't enter half the text
        for c in text:
            if c not in self.keys:
                raise ValueError("Don't know how to enter %r" % (c,))

        debug("MultiPress.enter_text: %r" % (text,))

        prev_key = None
        for c in text:
            key, n = self.keys[c]
            if prev_key == key:
                time.sleep(self.interletter_delay_secs)
            for _ in range(n):
                press(key, interpress_delay_secs=self.interpress_delay_secs)
            prev_key = key


def _parse_mapping_from_docstring(s):
    """
    >>> _parse_mapping_from_docstring(MultiPress.__doc__)['KEY_0']
    ' 0'
    >>> _parse_mapping_from_docstring(MultiPress.__doc__)['KEY_9']
    'wxyz9'
    """
    code = []
    in_mapping = False
    for line in s.split("\n"):
        if re.match(r"^            {", line):
            in_mapping = True
        if in_mapping:
            code.append(line)
        if re.match(r"^            }", line):
            break
    return eval("\n".join(code))  # pylint:disable=eval-used


def _letters_to_keys(keys_to_letters):
    """
    >>> sorted(_letters_to_keys({'KEY_1': '1', 'KEY_2': 'abc2'}).items())
    [('1', ('KEY_1', 1)),
     ('2', ('KEY_2', 4)),
     ('a', ('KEY_2', 1)),
     ('b', ('KEY_2', 2)),
     ('c', ('KEY_2', 3))]
    """
    out = {}
    for key, letters in keys_to_letters.items():
        for n, letter in enumerate(letters, start=1):
            out[letter] = (key, n)
    return out



from __future__ import annotations

import errno
import inspect
import os
import re
import typing
import warnings
from functools import lru_cache
from typing import Optional, overload, TypeAlias

import cv2
import numpy
import numpy.typing

from .logging import ddebug, debug, warn
from .types import Region


FrameT : TypeAlias = numpy.typing.NDArray[numpy.uint8]

# Anything that load_image can take:
ImageT : TypeAlias = numpy.typing.NDArray[numpy.uint8] | str


class Frame(numpy.ndarray):
    """A frame of video.

    A ``Frame`` is what you get from `stbt.get_frame` and `stbt.frames`. It is
    a subclass of `numpy.ndarray`, which is the type that OpenCV uses to
    represent images. Data is stored in 8-bit, 3 channel BGR format.

    In addition to the members inherited from `numpy.ndarray`, ``Frame``
    defines the following attributes:

    :ivar float time: The wall-clock time when this video-frame was captured,
        as number of seconds since the unix epoch (1970-01-01T00:00:00Z). This
        is the same format used by the Python standard library function
        `time.time`.
    :ivar int width: The width of the frame, in pixels.
    :ivar int height: The height of the frame, in pixels.
    :ivar Region region: A `Region` corresponding to the full size of the
        frame — that is, ``Region(0, 0, width, height)``.
    """
    def __new__(cls, array, dtype=None, order=None, time=None, _draw_sink=None):
        obj = numpy.asarray(array, dtype=dtype, order=order).view(cls)
        i = isinstance(array, Frame)
        if time is None and i:
            obj.time = array.time
        else:
            obj.time = time
        obj._draw_sink = _draw_sink or (i and array._draw_sink) or None
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.time = getattr(obj, 'time', None)  # pylint: disable=attribute-defined-outside-init
        self._draw_sink = getattr(obj, '_draw_sink', None)  # pylint: disable=attribute-defined-outside-init

    def __repr__(self):
        return "<Frame(time=%s)>" % (
            "None" if self.time is None else "%.3f" % self.time,)

    def __str__(self):
        return repr(self)

    @property
    def width(self) -> int:
        return self.shape[1]  # pylint:disable=unsubscriptable-object

    @property
    def height(self) -> int:
        return self.shape[0]  # pylint:disable=unsubscriptable-object

    @property
    def region(self) -> Region:
        return Region(0, 0, self.shape[1], self.shape[0])


class Image(numpy.ndarray):
    """An image, possibly loaded from disk.

    This is a subclass of `numpy.ndarray`, which is the type that OpenCV uses
    to represent images.

    In addition to the members inherited from `numpy.ndarray`, ``Image``
    defines the following attributes:

    :vartype filename: str or None
    :ivar filename: The filename that was given to `stbt.load_image`.

    :vartype absolute_filename: str or None
    :ivar absolute_filename: The absolute path resolved by `stbt.load_image`.

    :vartype relative_filename: str or None
    :ivar relative_filename: The path resolved by `stbt.load_image`, relative
        to the root of the test-pack git repo.

    Added in v32.
    """
    def __new__(cls, array, dtype=None, order=None,
                filename=None, absolute_filename=None):

        obj = numpy.asarray(array, dtype=dtype, order=order).view(cls)
        i = isinstance(array, Image)
        obj.filename = filename or (i and array.filename) or None
        obj.absolute_filename = (absolute_filename or
                                 (i and array.absolute_filename) or
                                 None)
        obj.relative_filename = _relative_filename(obj.absolute_filename)
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        # pylint: disable=attribute-defined-outside-init
        self.filename = getattr(obj, "filename", None)
        self.relative_filename = getattr(obj, "relative_filename", None)
        self.absolute_filename = getattr(obj, "absolute_filename", None)

    def __repr__(self):
        if len(self.shape) == 3:
            dimensions = "%dx%dx%d" % (
                self.shape[1], self.shape[0], self.shape[2])  # pylint:disable=unsubscriptable-object
        elif len(self.shape) == 2:
            dimensions = "%dx%d" % (self.shape[1], self.shape[0])  # pylint:disable=unsubscriptable-object
        else:
            return super().__repr__()
        if (self.relative_filename is None or
                self.relative_filename.startswith('../')):
            filename = self.absolute_filename
        else:
            filename = self.relative_filename
        return "<Image(filename=%r, dimensions=%s)>" % (filename, dimensions)

    def __str__(self):
        return repr(self)

    @property
    def width(self) -> int:
        return self.shape[1]  # pylint:disable=unsubscriptable-object

    @property
    def height(self) -> int:
        return self.shape[0]  # pylint:disable=unsubscriptable-object

    @property
    def region(self) -> Region:
        return Region(0, 0, self.shape[1], self.shape[0])


def _relative_filename(absolute_filename):
    """Returns filename relative to the test-pack root if inside the test-pack,
    or absolute path if outside the test-pack.
    """
    if absolute_filename is None:
        return None
    import stbt_core
    root = stbt_core.TEST_PACK_ROOT
    if root is None:
        return absolute_filename
    relpath = os.path.relpath(absolute_filename, root)
    if relpath.startswith(".."):
        return absolute_filename
    return relpath


def _frame_repr(frame):
    if frame is None:
        return "None"
    if isinstance(frame, (Image, Frame)):
        return repr(frame)
    if len(frame.shape) == 3:
        return "<%dx%dx%d>" % (frame.shape[1], frame.shape[0], frame.shape[2])
    else:
        return "<%dx%d>" % (frame.shape[1], frame.shape[0])


class Color:
    """A BGR color, optionally with an alpha (transparency) value.

    A Color can be created from an HTML-style hex string:

    >>> Color('#f77f00')
    Color('#f77f00')

    Or from Blue, Green, Red values in the range 0-255:

    >>> Color(0, 127, 247)
    Color('#f77f00')

    Note: When you specify the colors in this way, the BGR order is the
    opposite of the HTML-style RGB order. This is for compatibility with the
    way OpenCV stores colors.

    Any ``stbt`` APIs that take a ``Color`` will also accept a string or tuple
    in the above formats, so you don't need to construct a ``Color`` explicitly.
    """
    @overload
    def __init__(self, hexstring: str) -> None:
        ...
    @overload
    def __init__(self, blue: int, green: int, red: int,
                 alpha: Optional[int] = None) -> None:
        ...
    @overload
    def __init__(self, bgr: tuple[int, int, int]) -> None:
        ...
    @overload
    def __init__(self, bgra: tuple[int, int, int, int]) -> None:
        ...
    def __init__(self, *args,
                 hexstring: str = None,
                 blue: int = None, green: int = None, red: int = None,
                 alpha: int = None,
                 bgr: tuple[int, int, int] = None,
                 bgra: tuple[int, int, int, int] = None) -> None:

        self.array: numpy.ndarray  # BGR with shape (1, 1, 3) or BGRA (1, 1, 4)

        if len(args) == 1 and isinstance(args[0], Color):
            self.array = args[0].array

        elif len(args) == 1 and isinstance(args[0], str):
            self.array = Color._from_string(args[0])
        elif not args and hexstring is not None:
            self.array = Color._from_string(hexstring)

        elif (len(args) == 1 and
                isinstance(args[0], (list, tuple, numpy.ndarray)) and
                len(args[0]) in (3, 4)):
            self.array = Color._from_sequence(*args[0])
        elif len(args) in (3, 4):
            self.array = Color._from_sequence(*args)  # pylint:disable=no-value-for-parameter
        elif not args and bgr is not None:
            self.array = Color._from_sequence(*bgr)
        elif not args and bgra is not None:
            self.array = Color._from_sequence(*bgra)
        elif not args and all(x is not None for x in (blue, green, red)):
            self.array = Color._from_sequence(blue, green, red, alpha)

        else:
            raise TypeError("Color: __init__() expected a Color, '#rrggbb' "
                            "string, or 3 integers in Blue-Green-Red order.")

        self.array.flags.writeable = False

        self.hexstring = (
            "#{0:02x}{1:02x}{2:02x}{3}".format(
                self.array[0][0][2], self.array[0][0][1], self.array[0][0][0],
                ("" if self.array.shape[2] == 3
                 else f"{self.array[0][0][3]:02x}")))

    @staticmethod
    def _from_string(s):
        m = re.match(r"^#?([0-9a-fA-F]{8}|[0-9a-fA-F]{6}|[0-9a-fA-F]{3})$", s)
        if m:
            s = m.group(1)
            if len(s) == 8:
                r = s[0:2]
                g = s[2:4]
                b = s[4:6]
                a = s[6:8]
            elif len(s) == 6:
                r = s[0:2]
                g = s[2:4]
                b = s[4:6]
                a = None
            elif len(s) == 3:
                r = s[0] * 2
                g = s[1] * 2
                b = s[2] * 2
                a = None
            else:
                assert False, f"unreachable: {s}"
        else:
            raise ValueError(
                f"Invalid color string. Expected hexadesimal digits in the "
                f"format '#rrggbb'; got {s!r}")
        return Color._from_sequence(b, g, r, a)

    @staticmethod
    def _from_sequence(b, g, r, a=None):
        channels = [b, g, r]
        if a is not None:
            channels.append(a)
        out = []
        for x in channels:
            if isinstance(x, str):
                x = int(x, 16)
            else:
                x = int(x)
            if not 0 <= x <= 255:
                raise ValueError(f"Color: __init__ expected a value between 0 "
                                 f"and 255: Got {x}")
            out.append(x)
        return (numpy.asarray(out, dtype=numpy.uint8)
                     .reshape((1, 1, len(channels))))

    def __repr__(self) -> str:
        return f"Color('{self.hexstring}')"

    def __eq__(self, other: Color) -> bool:
        return isinstance(other, Color) and self.hexstring == other.hexstring

    def __hash__(self) -> int:
        return hash(self.hexstring)


ColorT : TypeAlias = Color | str | tuple[int, int, int]


def crop(frame: FrameT, region: Region) -> FrameT:
    """Returns an image containing the specified region of ``frame``.

    :type frame: `stbt.Frame` or `numpy.ndarray`
    :param frame: An image in OpenCV format (for example as returned by
      `frames`, `get_frame` and `load_image`, or the ``frame`` parameter of
      `MatchResult`).

    :type Region region: The region to crop.

    :returns: An OpenCV image (`numpy.ndarray`) containing the specified region
      of the source frame. This is a view onto the original data, so if you
      want to modify the cropped image call its ``copy()`` method first.
    """
    r = _validate_region(frame, region)
    return frame[r.y:r.bottom, r.x:r.right]


def _validate_region(frame, region):
    if region is None:
        raise TypeError(
            "'region=None' means an empty region. To analyse the entire "
            "frame use 'region=Region.ALL' (which is the default)")
    f = _image_region(frame)
    r = Region.intersect(f, region)
    if r is None:
        raise ValueError("%r doesn't overlap with the frame dimensions %ix%i"
                         % (region, frame.shape[1], frame.shape[0]))
    return r


def _image_region(image):
    s = image.shape
    return Region(0, 0, s[1], s[0])


@typing.overload
def load_image(filename: ImageT) -> Image:
    ...


@typing.overload
def load_image(filename: ImageT, flags: int) -> Image:
    ...


@typing.overload
def load_image(
        filename: ImageT, *, color_channels: int | tuple[int, ...]) -> Image:
    ...


def load_image(filename, flags=None, color_channels=None) -> Image:
    """Find & read an image from disk.

    If given a relative filename, this will search in the directory of the
    Python file that called ``load_image``, then in the directory of that
    file's caller, and so on, until it finds the file. This allows you to use
    ``load_image`` in a helper function that takes a filename from its caller.

    Finally this will search in the current working directory. This allows
    loading an image that you had previously saved to disk during the same
    test run.

    This is the same search algorithm used by `stbt.match` and similar
    functions.

    :param str filename: A relative or absolute filename.

    :param flags: Flags to pass to :ocv:pyfunc:`cv2.imread`. Deprecated; use
      ``color_channels`` instead.

    :param tuple[int] color_channels: Tuple of acceptable numbers of color
      channels for the output image: 1 for grayscale, 3 for color, and 4 for
      color with an alpha (transparency) channel. For example,
      ``color_channels=(3, 4)`` will accept color images with or without an
      alpha channel. Defaults to ``(3, 4)``.

      If the image doesn't match the specified ``color_channels`` it will be
      converted to the specified format.

    :rtype: stbt.Image
    :returns: An image in OpenCV format — that is, a `numpy.ndarray` of 8-bit
        values. With the default ``color_channels`` parameter this will be 3
        channels BGR, or 4 channels BGRA if the file has transparent pixels.
    :raises: `IOError` if the specified path doesn't exist or isn't a valid
        image file.

    * Changed in v32: Return type is now `stbt.Image`, which is a
      `numpy.ndarray` sub-class with additional attributes ``filename``,
      ``relative_filename`` and ``absolute_filename``.
    * Changed in v32: Allows passing an image (`numpy.ndarray` or `stbt.Image`)
      instead of a string, in which case this function returns the given image.
    * Changed in v33: Added the ``color_channels`` parameter and deprecated
      ``flags``. The image will always be converted to the format specified by
      ``color_channels`` (previously it was only converted to the format
      specified by ``flags`` if it was given as a filename, not as a
      `stbt.Image` or numpy array). The returned numpy array is read-only.
    """
    if flags is not None:
        # Backwards compatibility
        if color_channels is not None:
            raise Exception(
                "flags cannot be specified at the same time as color_channels")
        if flags == cv2.IMREAD_GRAYSCALE:
            color_channels = (1,)
        elif flags == cv2.IMREAD_COLOR:
            color_channels = (3,)
        elif flags == cv2.IMREAD_UNCHANGED:
            color_channels = (1, 3, 4)
        else:
            raise Exception("Unsupported imread flags %s" % flags)
        warnings.warn(
            "load_image: flags=%s argument is deprecated. Use "
            "color_channels=%r instead" % (flags, color_channels),
            DeprecationWarning, stacklevel=2)

    if color_channels is None:
        color_channels = (3, 4)
    elif isinstance(color_channels, int):
        color_channels = (color_channels,)

    obj = filename
    if isinstance(obj, Image):
        filename = obj.filename
        absolute_filename = obj.absolute_filename
        img = _convert_color(obj, color_channels, absolute_filename)
    elif isinstance(obj, numpy.ndarray):
        filename = None
        absolute_filename = None
        img = _convert_color(obj, color_channels, absolute_filename)
        if img.shape[2] == 4:
            # Normalise transparency channel to either 0 or 255, for stbt.match
            img = img.copy()
            chan = img[:, :, 3]
            chan[chan < 255] = 0
    elif isinstance(filename, str):
        absolute_filename = find_file(filename)
        img = _imread(absolute_filename, color_channels)
    else:
        raise TypeError("load_image requires a filename or Image")

    if not isinstance(img, Image):
        img = Image(img, filename=filename, absolute_filename=absolute_filename)
    return img


@lru_cache(maxsize=5)
def _imread(absolute_filename, color_channels):
    if color_channels == (3,):
        flags = cv2.IMREAD_COLOR
    elif color_channels == (1,):
        flags = cv2.IMREAD_GRAYSCALE
    else:
        flags = cv2.IMREAD_UNCHANGED
    img = cv2.imread(absolute_filename, flags)
    if img is None:
        raise IOError("Failed to load image: %s" % absolute_filename)
    img = _convert_color(img, color_channels, absolute_filename)
    if img.shape[2] == 4:
        # Normalise transparency channel to either 0 or 255, for stbt.match
        chan = img[:, :, 3]
        chan[chan < 255] = 0
    img.flags.writeable = False
    return img


def _convert_color(img, color_channels, absolute_filename):
    if len(img.shape) not in [2, 3]:
        raise ValueError(
            "Invalid shape for image: %r. Shape must have 2 or 3 elements" %
            (img.shape,))

    if img.dtype == numpy.uint16:
        warn("Image %s has 16 bits per channel. Converting to 8 bits."
             % _filename_repr(absolute_filename))
        img = cv2.convertScaleAbs(img, alpha=1.0 / 256)
    elif img.dtype != numpy.uint8:
        raise ValueError("Image %s must be 8-bits per channel (got %s)"
                         % (_filename_repr(absolute_filename), img.dtype))

    if len(img.shape) == 2:
        img = img.reshape(img.shape + (1,))
    c = img.shape[2]

    if c == 1:
        if 1 in color_channels:
            pass
        elif 3 in color_channels:
            img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
        elif 4 in color_channels:
            img = numpy.concatenate(
                (cv2.cvtColor(img, cv2.COLOR_GRAY2BGR),
                 numpy.full(img.shape[:2] + (1,), 255, dtype=img.dtype)),
                axis=2)
        else:
            raise ValueError(
                "Can only convert 1 channel image to 1, 3 or 4 channels")
    elif c == 3:
        if 3 in color_channels:
            pass
        elif 4 in color_channels:
            img = numpy.concatenate(
                (img, numpy.full(img.shape[:2] + (1,), 255, dtype=img.dtype)),
                axis=2)
        elif 1 in color_channels:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
            img = img.reshape(img.shape + (1,))
        else:
            raise ValueError(
                "Can only convert 3 channel image to 1, 3 or 4 channels")
    elif c == 4:
        if 4 in color_channels:
            # Remove alpha channel if it's 100% opaque
            if 3 in color_channels and numpy.all(img[:, :, 3] == 255):
                img = img[:, :, :3]
        elif 3 in color_channels:
            img = img[:, :, :3]
        elif 1 in color_channels:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
            img = img.reshape(img.shape + (1,))
        else:
            raise ValueError(
                "Can only convert 3 channel image to 1, 3 or 4 channels")
    else:
        raise ValueError(
            "load_image can only handle images with 1, 3 or 4 color channels. "
            "%s has %i channels" % (_filename_repr(absolute_filename), c))

    assert img.shape[2] in color_channels
    return img


def _filename_repr(absolute_filename):
    if absolute_filename is None:
        return "<Image>"
    else:
        return repr(_relative_filename(absolute_filename))


def save_frame(image: FrameT, filename: str):
    """Saves an OpenCV image to the specified file.

    Takes an image obtained from `get_frame` or from the `screenshot`
    property of `MatchTimeout` or `MotionTimeout`.
    """
    cv2.imwrite(filename, image)


def pixel_bounding_box(img):
    """
    Find the smallest region that contains all the non-zero pixels in an image.

    >>> pixel_bounding_box(numpy.array([[0]], dtype=numpy.uint8))
    >>> pixel_bounding_box(numpy.array([[1]], dtype=numpy.uint8))
    Region(x=0, y=0, right=1, bottom=1)
    >>> pixel_bounding_box(numpy.array([[[1]]], dtype=numpy.uint8))
    Region(x=0, y=0, right=1, bottom=1)
    >>> pixel_bounding_box(numpy.array([[[1, 1, 1]]], dtype=numpy.uint8))
    Region(x=0, y=0, right=1, bottom=1)
    >>> a = numpy.array([
    ...     [0, 0, 0, 0],
    ...     [0, 1, 1, 1],
    ...     [0, 1, 1, 1],
    ...     [0, 0, 0, 0],
    ... ], dtype=numpy.uint8)
    >>> pixel_bounding_box(a)
    Region(x=1, y=1, right=4, bottom=3)
    >>> pixel_bounding_box(numpy.stack([
    ...     numpy.zeros((4, 4), dtype=numpy.uint8),
    ...     numpy.zeros((4, 4), dtype=numpy.uint8),
    ...     a],
    ...     axis=-1))
    Region(x=1, y=1, right=4, bottom=3)
    >>> pixel_bounding_box(numpy.array([
    ...     [0, 0, 0, 0, 0, 0],
    ...     [0, 0, 0, 1, 0, 0],
    ...     [0, 1, 0, 0, 0, 0],
    ...     [0, 0, 0, 0, 1, 0],
    ...     [0, 0, 0, 0, 0, 0],
    ...     [0, 0, 1, 0, 0, 0],
    ...     [0, 0, 0, 0, 0, 0]
    ... ], dtype=numpy.uint8))
    Region(x=1, y=1, right=5, bottom=6)
    """
    if len(img.shape) == 2 or len(img.shape) == 3 and img.shape[2] == 1:
        pass
    elif len(img.shape) == 3 and img.shape[2] == 3:
        img = img.max(axis=2)
    else:
        raise ValueError("Single-channel or 3-channel (BGR) image required. "
                         "Provided image has shape %r" % (img.shape,))
    rect = cv2.boundingRect(img)
    if rect[2] == 0 or rect[3] == 0:
        return None
    else:
        return Region(*rect)


def find_file(filename: str) -> str:
    """Searches for the given filename relative to the directory of the caller.

    When Stb-tester runs a test, the "current working directory" is not the
    same as the directory of the test-pack git checkout. If you want to read
    a file that's committed to git (for example a CSV file with data that your
    test needs) you can use this function to find it. For example::

        f = open(stbt.find_file("my_data.csv"))

    If the file is not found in the directory of the Python file that called
    ``find_file``, this will continue searching in the directory of that
    function's caller, and so on, until it finds the file. This allows you to
    use ``find_file`` in a helper function that takes a filename from its
    caller.

    This is the same algorithm used by `load_image`.

    :param str filename: A relative filename.

    :rtype: str
    :returns: Absolute filename.
    :raises: `FileNotFoundError` if the file can't be found.

    Added in v33.
    """
    if os.path.isabs(filename):
        if os.path.isfile(filename):
            return filename

    else:
        caller = inspect.currentframe()
        try:
            caller = caller.f_back  # skip this frame (find_file)
            while caller:
                caller_dir = os.path.abspath(
                    os.path.dirname(inspect.getframeinfo(caller).filename))
                caller_path = os.path.join(caller_dir, filename)
                if os.path.isfile(caller_path):
                    ddebug("Resolved relative path %r to %r" % (
                        filename, caller_path))
                    return caller_path
                caller = caller.f_back
        finally:
            # Avoid circular references between stack frame objects and
            # themselves for more deterministic GC. See
            # https://docs.python.org/3.10/library/inspect.html#the-interpreter-stack
            del caller

        # Fall back to image from cwd, to allow loading an image saved
        # previously during the same test-run.
        if os.path.isfile(filename):
            abspath = os.path.abspath(filename)
            ddebug("Resolved relative path %r to %r" % (filename, abspath))
            return abspath

    raise FileNotFoundError(errno.ENOENT, "No such file", filename)


def limit_time(frames, duration_secs):
    """
    Adapts a frame iterator such that it will return EOS after `duration_secs`
    worth of video has been read.
    """
    import time
    end_time = time.time() + duration_secs
    for frame in frames:
        if frame.time > end_time:
            debug("timed out: %.3f > %.3f" % (frame.time, end_time))
            break
        else:
            yield frame



from __future__ import annotations

import ctypes
import os
import platform

import numpy


def _find_file(path, root=os.path.dirname(os.path.abspath(__file__))):
    return os.path.join(root, path)


try:
    _libstbt = ctypes.CDLL(_find_file(f"libstbt.{platform.machine()}.so"))
except OSError:
    raise ImportError("Failed to load libstbt.so")


class _SqdiffResult(ctypes.Structure):
    _fields_ = [("total", ctypes.c_uint64),
                ("count", ctypes.c_uint32)]


# SqdiffResult sqdiff(const uint8_t *t, uint16_t t_stride,
#                     const uint8_t *f, uint16_t f_stride,
#                     uint16_t width_px, uint16_t height_px,
#                     int color_depth)

_libstbt.sqdiff.restype = _SqdiffResult
_libstbt.sqdiff.argtypes = [
    ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint16,
    ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint16,
    ctypes.c_uint16, ctypes.c_uint16,
    ctypes.c_int
]

# void threshold_diff_bgr(
#     uint8_t *out,
#     const uint8_t* a, uint16_t line_stride_a,
#     uint8_t* b, uint16_t line_stride_b,
#     uint32_t threshold_sq,
#     uint16_t width_px, uint16_t height_px
# )
_libstbt.threshold_diff_bgr.argtypes = [
    ctypes.POINTER(ctypes.c_uint8),
    ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint16,
    ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint16,
    ctypes.c_uint32,
    ctypes.c_uint16, ctypes.c_uint16
]

PIXEL_DEPTH_BGR = 1
PIXEL_DEPTH_BGRx = 2
PIXEL_DEPTH_BGRA = 3

COLOR_DEPTH_LOOKUP = {
    (3, 3): PIXEL_DEPTH_BGR,
    (4, 3): PIXEL_DEPTH_BGRx,
    (4, 4): PIXEL_DEPTH_BGRA,
}


def sqdiff(template, frame):
    if template.dtype != numpy.uint8 or frame.dtype != numpy.uint8:
        raise NotImplementedError("dtype must be uint8")

    if frame.strides[2] != 1 or template.strides[2] != 1 or \
            frame.strides[1] != 3:
        raise NotImplementedError("Pixel data must be contiguous")

    color_depth = COLOR_DEPTH_LOOKUP[(template.strides[1], template.shape[2])]

    t = template.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8))
    f = frame.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8))

    out = _libstbt.sqdiff(t, template.strides[0],
                          f, frame.strides[0],
                          template.shape[1], template.shape[0], color_depth)
    return out.total, out.count


def threshold_diff_bgr(
        a: numpy.ndarray[numpy.uint8], b: numpy.ndarray[numpy.uint8],
        threshold: int) -> numpy.ndarray[numpy.uint8]:
    if a.dtype != numpy.uint8 or b.dtype != numpy.uint8:
        raise NotImplementedError("dtype must be uint8")

    if b.strides[2] != 1 or a.strides[2] != 1 or \
            b.strides[1] != 3 or a.strides[1] != 3:
        raise NotImplementedError("Pixel data must be contiguous")

    out = numpy.empty(a.shape[:2], dtype=numpy.uint8)

    a_array = a.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8))
    b_array = b.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8))
    out_array = out.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8))

    _libstbt.threshold_diff_bgr(
        out_array, a_array, a.strides[0], b_array, b.strides[0],
        threshold, a.shape[1], a.shape[0])
    return out



# -*- coding: utf-8 -*-

"""Python module to control Android devices via `ADB`_ from Stb-tester scripts.

Copyright © 2017-2022 Stb-tester.com Ltd.
License: LGPL v2.1 or (at your option) any later version (see
https://github.com/stb-tester/stb-tester/blob/master/LICENSE for details).

.. _ADB: https://developer.android.com/studio/command-line/adb.html
"""

import os
import re
import shutil
import subprocess
import threading
import time
from collections import namedtuple
from contextlib import contextmanager
from logging import getLogger

from enum import Enum

from _stbt.config import ConfigurationError, get_config


__all__ = [
    "adb",
    "AdbDevice",
    "AdbError",
    "CoordinateSystem",
]


logger = getLogger("stbt.android")


class CoordinateSystem(Enum):
    """How to translate coordinates from the video-frames processed by your
    test script, to the coordinates expected by ADB for tap & swipe events.

    When you tell ADB to send a tap or swipe event you must give it x & y
    coordinates. The coordinate system matches the orientation of the device.
    We'll call these the physical coordinates. For example my phone's display
    is Full HD (1080 pixels x 1920 pixels):

    ===================  ======  ======  ======================================
    Logical orientation  x       y       (0, 0) is
    ===================  ======  ======  ======================================
    Portrait             0-1080  0-1920  top-left
    Landscape            0-1920  0-1080  top-left (i.e. the top-right corner if
                                         the phone is physically in portrait)
    ===================  ======  ======  ======================================

    Note that an app can force the device to think it's in portrait orientation
    even if you're physically holding the device in landscape orientation. We
    call this the "logical orientation". Similarly, an app can force the device
    into logical landscape orientation even if you have switched off auto-rotate
    in the device's settings menu.

    The video-frames that you analyse in your test scripts may not match the
    physical resolution of the device. For example if you're using a camera
    pointed at the device's screen, the resolution of your screenshots (even
    after geometric correction by the `Stb-tester CAMERA`_) will depend on the
    resolution of the camera, not of the device under test.

    You can give `AdbDevice.tap` and `AdbDevice.swipe` the coordinates from
    your video-frame (for example, you can pass in ``stbt.match(...).region``)
    and the CoordinateSystem will ensure that the coordinates are converted to
    physical coordinates in the appropriate way before passing them on to ADB.

    .. _Stb-tester CAMERA: https://stb-tester.com/stb-tester-camera
    """

    ADB_NATIVE = 0
    """Frames are captured via ADB screenshot.

    Frames will be in the same orientation & resolution as the physical device.
    """

    ADB_720P = 1
    """Frames are captured via ADB screenshot and then scaled to 720p.

    Frames will be in the same orientation as the physical device, but scaled.
    """

    HDMI_720P = 2
    """
    Frames are captured via HDMI and scaled to 720p.

    Frames will always be in landscape orientation. If a mobile device is in
    portrait orientation, you'll get black bars to the left & right. If the
    device is in landscape orientation, the frame will match what you see on
    the device (this assumes that the device's physical aspect ratio matches
    the aspect ratio of HDMI).
    """

    CAMERA_720P = 3
    """Frames are captured from an `Stb-tester CAMERA`_ pointing at the
    device's screen.

    The camera & device must both be physically in landscape orientation.

    Frames will always be in landscape orientation; if the device is in logical
    portrait orientation the image will appear rotated 90° anti-clockwise.
    """


def adb(args, **subprocess_kwargs) -> subprocess.CompletedProcess:
    """Send commands to an Android device using `ADB`_.

    This is a convenience function. It will construct an `AdbDevice` with the
    default parameters (taken from your config files) and call `AdbDevice.adb`
    with the parameters given here.

    .. _ADB: https://developer.android.com/studio/command-line/adb.html
    """
    return AdbDevice().adb(args, **subprocess_kwargs)


class AdbDevice():
    """Send commands to an Android device using `ADB`_.

    Default values for each parameter can be specified in your "stbt.conf"
    config file under the "[android]" section.

    :param string address:
        IP address (if using Network ADB) or serial number (if connected via
        USB) of the Android device. You can get the serial number by running
        ``adb devices -l``. If not specified, this is read from
        "device_under_test.ip_address" in your :ref:`node-specific-config`.
    :param string adb_server:
        The ADB server (that is, the PC connected to the Android device).
        Defaults to localhost.
    :param string adb_binary:
        The path to the ADB client executable. Defaults to "adb".
    :param bool tcpip:
        The ADB server communicates with the Android device via TCP/IP, not
        USB. This requires that you have enabled Network ADB access on the
        device. Defaults to True if ``address`` is an IP address, False
        otherwise.

    .. _ADB: https://developer.android.com/studio/command-line/adb.html
    """

    def __init__(self, address=None, adb_server=None, adb_binary=None,
                 tcpip=None, coordinate_system=None):

        self.address = address or get_config("device_under_test", "ip_address",
                                             default=None)
        self.adb_server = adb_server or get_config("android", "adb_server",
                                                   default=None)
        self.adb_binary = adb_binary or get_config("android", "adb_binary",
                                                   default="adb")

        if tcpip is None:
            tcpip = get_config("android", "tcpip", default=None, type_=bool)
        if tcpip is None:
            tcpip = _is_ip_address(self.address)
        self.tcpip = tcpip

        if self.tcpip and not self.address:
            raise ConfigurationError('AdbDevice: If "tcpip=True" '
                                     'you must specify "address"')

        if self.tcpip and ":" not in self.address:
            self.address = self.address + ":5555"

        self.coordinate_system = coordinate_system or get_config(
            "android", "coordinate_system", default=CoordinateSystem.HDMI_720P,
            type_=CoordinateSystem)

        self._setup_adb_key()

        if self.tcpip:
            self._connect(timeout=60)

    def _setup_adb_key(self):
        if os.path.exists(os.path.join(os.environ["HOME"], ".android/adbkey")):
            return
        import stbt_core
        root = stbt_core.TEST_PACK_ROOT
        if root is None:
            return
        if not os.path.exists(os.path.join(root, "config/android/adbkey")):
            raise ConfigurationError(
                "To run adb on the Stb-tester Node you must generate an ADB "
                "host key by running 'adb keygen config/android/adbkey' and "
                "commit the 'config/android' directory to git.")
        shutil.copytree(os.path.join(root, "config/android"),
                        os.path.join(os.environ["HOME"], ".android"))

    def adb(self, args, *, timeout=None, **subprocess_kwargs) \
            -> subprocess.CompletedProcess:
        """Run any ADB command.

        For example, the following code will use "adb shell am start" to launch
        an app on the device::

            d = AdbDevice(...)
            d.adb(["shell", "am", "start", "-S",
                   "com.example.myapp/com.example.myapp.MainActivity"])

        Any keyword arguments are passed on to `subprocess.run`.

        :returns: `subprocess.CompletedProcess` from `subprocess.run`.
        :raises: `subprocess.CalledProcessError` if ``check`` is true and
            the adb process returns a non-zero exit status.
        :raises: `subprocess.TimeoutExpired` if ``timeout`` is specified and
            the adb process doesn't finish within that number of seconds.
        :raises: `stbt.android.AdbError` if ``adb connect`` fails.
        """
        if self.tcpip:
            self._connect(timeout)
        return self._adb(args, timeout=timeout, **subprocess_kwargs)

    def devices(self) -> str:
        """Output of ``adb devices -l``."""
        return self._adb(["devices", "-l"], verbose=False, timeout=5,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.STDOUT,
                         encoding="utf-8").stdout

    def get_frame(self, coordinate_system=None) -> "stbt.Frame":
        """Take a screenshot using ADB.

        If you are capturing video from the Android device via another method
        (namely, HDMI capture) sometimes it can be useful to capture a frame
        via ADB for debugging. This function will manipulate the ADB screenshot
        (scale and/or rotate it) to match the screenshots from your main
        video-capture method as closely as possible.

        :returns: A `stbt.Frame`, that is, an image in OpenCV format. Note that
            the ``time`` attribute won't be very accurate (probably to <0.5s or
            so).
        """

        import cv2
        import numpy
        from _stbt.imgutils import Frame

        if coordinate_system is None:
            coordinate_system = self.coordinate_system

        for attempt in range(1, 4):
            timestamp = time.time()
            data = self.adb(["shell", "screencap", "-p"],
                            timeout=60, capture_output=True) \
                       .stdout.replace(b"\r\n", b"\n")
            img = cv2.imdecode(
                numpy.asarray(bytearray(data), dtype=numpy.uint8),
                cv2.IMREAD_COLOR)
            if img is None:
                logger.warning(
                    "AdbDevice.get_frame: Failed to get screenshot "
                    "via ADB (attempt %d/3)\n"
                    "Length of data: %d", attempt, len(data))
            else:
                break
        else:
            raise AdbError(
                "Failed to capture screenshot from android device")

        img = _resize(img, coordinate_system)
        return Frame(img, time=timestamp)

    def press(self, key) -> None:
        """Send a keypress.

        :param str key: An Android keycode as listed in
          <https://developer.android.com/reference/android/view/KeyEvent.html>.
          Also accepts standard Stb-tester key names like "KEY_HOME" and
          "KEY_BACK".
        """
        # "adb shell input keyevent xxx" always returns success, so we need to
        # validate key names.
        if key in _KEYCODE_MAPPINGS:  # pylint:disable=consider-using-get
            key = _KEYCODE_MAPPINGS[key]  # Map Stb-tester names to Android ones
        if key not in _ANDROID_KEYCODES:
            raise ValueError("Unknown key code %r" % (key,))
        logger.info("AdbDevice.press(%r)", key)
        self.adb(["shell", "input", "keyevent", key], timeout=10)

    def swipe(self, start_position, end_position) -> None:
        """Swipe from one point to another point.

        :param start_position:
            A `stbt.Region` or (x, y) tuple of coordinates at which to start.
        :param end_position:
            A `stbt.Region` or (x, y) tuple of coordinates at which to stop.

        Example::

          d.swipe((100, 100), (100, 400))

        """
        x1, y1 = _centre_point(start_position)
        x2, y2 = _centre_point(end_position)
        logger.info("AdbDevice.swipe((%d,%d), (%d,%d))", x1, y1, x2, y2)

        x1, y1 = self._to_native_coordinates(x1, y1)
        x2, y2 = self._to_native_coordinates(x2, y2)
        command = ["shell", "input", "swipe",
                   str(x1), str(y1), str(x2), str(y2)]
        self.adb(command, timeout=10)

    def tap(self, position) -> None:
        """Tap on a particular location.

        :param position: A `stbt.Region`, or an (x,y) tuple.

        Example::

            d.tap((100, 20))
            d.tap(stbt.match(...).region)

        """
        x, y = _centre_point(position)
        logger.info("AdbDevice.tap((%d,%d))", x, y)

        x, y = self._to_native_coordinates(x, y)
        self.adb(["shell", "input", "tap", str(x), str(y)], timeout=10)

    @contextmanager
    def logcat(self, filename="logcat.log", logcat_args=None):
        """Run ``adb logcat`` and stream the logs to ``filename``.

        This is a context manager. You can use it as a decorator on your
        test-case functions, and ``adb logcat`` will run for the duration
        of the decorated function::

            adb = stbt.android.AdbDevice()

            @adb.logcat()
            def test_launching_my_androidtv_app():
                ...

        :param str filename:
            Where the logs are written.

        :param list logcat_args:
            Optional arguments to pass on to ``adb logcat``, such as filter
            expressions. For example:
            ``logcat_args=["ActivityManager:I", "MyApp:D", "*:S"]``.
            See the `logcat documentation
            <https://developer.android.com/studio/command-line/logcat#options>`__.
        """
        collector = _LogcatCollector(filename, self, logcat_args)
        self.adb(["logcat", "--clear"])
        collector.run_in_background()
        try:
            yield
        finally:
            collector.stop()

    def _adb(self, args, verbose=True, **kwargs):
        _command = self._build_adb_command() + args
        if verbose:
            logger.debug("AdbDevice.adb: About to run command: %r", _command)
        return subprocess.run(_command, **kwargs)  # pylint:disable=subprocess-run-check

    def _Popen(self, args, **kwargs):
        """Like `AdbDevice.adb`, but runs `subprocess.Popen` instead of
        `subprocess.run`.
        """
        if self.tcpip:
            self._connect()
        _command = self._build_adb_command() + args
        logger.debug("AdbDevice._Popen: About to run command: %r", _command)
        return subprocess.Popen(_command, **kwargs)

    def _build_adb_command(self):
        _command = []
        _command += [self.adb_binary]
        if self.adb_server:
            _command += ["-H", self.adb_server]
        if self.address:
            _command += ["-s", self.address]
        return _command

    def _connect(self, timeout=None):
        if self.address in self.devices():
            return

        if timeout is None:
            timeout = 60

        # "adb connect" always returns success; we have to parse the output
        # which looks like "connected to 192.168.2.163:5555" or
        # "already connected to 192.168.2.163:5555" or
        # "unable to connect to 192.168.2.100:5555".
        output = self._adb(["connect", self.address], timeout=timeout,
                           stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                           encoding="utf-8") \
                     .stdout
        if ("connected to %s" % self.address) not in output:
            logger.debug("%s", output)
            raise AdbError("adb connect %s failed" % self.address,
                           output=output, devices=self.devices())
        time.sleep(2)

    def _to_native_coordinates(self, x, y):
        if self.coordinate_system == CoordinateSystem.ADB_NATIVE:
            return x, y
        else:
            return _to_native_coordinates(
                x, y, self.coordinate_system, self._get_display_dimensions())

    def _get_display_dimensions(self):
        return _parse_display_dimensions(
            self.adb(["shell", "dumpsys", "window"],
                     timeout=10, capture_output=True).stdout)


class AdbError(Exception):
    """Exception raised by `AdbDevice.adb`.

    :ivar int returncode: Exit status of the adb command.
    :ivar list cmd: The command that failed, as given to `AdbDevice.adb`.
    :ivar str output: The output from adb.
    :ivar str devices: The output from "adb devices -l" (useful for
        debugging connection errors).
    """

    def __init__(self, message, returncode=None, cmd=None, output=None,
                 devices=None):
        super().__init__()
        self.message = message
        self.returncode = returncode
        self.cmd = cmd
        self.output = output
        self.devices = devices

    def __str__(self):
        return self.message


class _LogcatCollector():
    def __init__(self, filename, adb_device, logcat_args=None):
        self.filename = filename
        self.adb_device = adb_device
        self.logcat_args = logcat_args or []
        self.stopping = False
        self.thread = None
        self.process = None

    def run_in_background(self):
        thread = threading.Thread(target=self.run)
        thread.daemon = True
        self.thread = thread
        thread.start()

    def run(self):
        with open(self.filename, "wb", 0) as f:
            while True:
                # Re-run "adb logcat" in case of disconnections. I assume this
                # means the device-under-test crashed/rebooted so we don't need
                # to run "logcat --clear" again. If this assumption is wrong
                # we'll end up with some duplicate log lines. :shrug:
                self.process = self.adb_device._Popen(
                    ['logcat'] + self.logcat_args, stdout=f)
                self.process.wait()
                if self.stopping:
                    return
                logger.warning("logcat: Lost adb connection")
                time.sleep(1)

    def stop(self):
        self.stopping = True
        if self.process:
            self.process.terminate()
            try:
                self.process.wait(timeout=10)
            except subprocess.TimeoutExpired:
                self.process.kill()


# https://developer.android.com/reference/android/view/KeyEvent.html
#
# I generated this list with:
#
#   curl https://android.googlesource.com/platform/frameworks/base/+/master/core/java/android/view/KeyEvent.java?format=TEXT |
#   base64 -d | grep -o 'KEYCODE_[A-Z0-9_]*' | sort | uniq |
#   grep -v -e KEYCODE_UNKNOWN -e 'KEYCODE_$' | sed 's/^.*$/    "&",/'
_ANDROID_KEYCODES = [
    "KEYCODE_0",
    "KEYCODE_1",
    "KEYCODE_11",
    "KEYCODE_12",
    "KEYCODE_2",
    "KEYCODE_3",
    "KEYCODE_3D_MODE",
    "KEYCODE_4",
    "KEYCODE_5",
    "KEYCODE_6",
    "KEYCODE_7",
    "KEYCODE_8",
    "KEYCODE_9",
    "KEYCODE_A",
    "KEYCODE_ALL_APPS",
    "KEYCODE_ALT_LEFT",
    "KEYCODE_ALT_RIGHT",
    "KEYCODE_APOSTROPHE",
    "KEYCODE_APP_SWITCH",
    "KEYCODE_ASSIST",
    "KEYCODE_AT",
    "KEYCODE_AVR_INPUT",
    "KEYCODE_AVR_POWER",
    "KEYCODE_B",
    "KEYCODE_BACK",
    "KEYCODE_BACKSLASH",
    "KEYCODE_BOOKMARK",
    "KEYCODE_BREAK",
    "KEYCODE_BRIGHTNESS_DOWN",
    "KEYCODE_BRIGHTNESS_UP",
    "KEYCODE_BUTTON_1",
    "KEYCODE_BUTTON_10",
    "KEYCODE_BUTTON_11",
    "KEYCODE_BUTTON_12",
    "KEYCODE_BUTTON_13",
    "KEYCODE_BUTTON_14",
    "KEYCODE_BUTTON_15",
    "KEYCODE_BUTTON_16",
    "KEYCODE_BUTTON_2",
    "KEYCODE_BUTTON_3",
    "KEYCODE_BUTTON_4",
    "KEYCODE_BUTTON_5",
    "KEYCODE_BUTTON_6",
    "KEYCODE_BUTTON_7",
    "KEYCODE_BUTTON_8",
    "KEYCODE_BUTTON_9",
    "KEYCODE_BUTTON_A",
    "KEYCODE_BUTTON_B",
    "KEYCODE_BUTTON_C",
    "KEYCODE_BUTTON_L1",
    "KEYCODE_BUTTON_L2",
    "KEYCODE_BUTTON_MODE",
    "KEYCODE_BUTTON_R1",
    "KEYCODE_BUTTON_R2",
    "KEYCODE_BUTTON_SELECT",
    "KEYCODE_BUTTON_START",
    "KEYCODE_BUTTON_THUMBL",
    "KEYCODE_BUTTON_THUMBR",
    "KEYCODE_BUTTON_X",
    "KEYCODE_BUTTON_Y",
    "KEYCODE_BUTTON_Z",
    "KEYCODE_C",
    "KEYCODE_CALCULATOR",
    "KEYCODE_CALENDAR",
    "KEYCODE_CALL",
    "KEYCODE_CAMERA",
    "KEYCODE_CAPS_LOCK",
    "KEYCODE_CAPTIONS",
    "KEYCODE_CHANNEL_DOWN",
    "KEYCODE_CHANNEL_UP",
    "KEYCODE_CLEAR",
    "KEYCODE_COMMA",
    "KEYCODE_CONTACTS",
    "KEYCODE_COPY",
    "KEYCODE_CTRL_LEFT",
    "KEYCODE_CTRL_RIGHT",
    "KEYCODE_CUT",
    "KEYCODE_D",
    "KEYCODE_DEL",
    "KEYCODE_DPAD_CENTER",
    "KEYCODE_DPAD_DOWN",
    "KEYCODE_DPAD_DOWN_LEFT",
    "KEYCODE_DPAD_DOWN_RIGHT",
    "KEYCODE_DPAD_LEFT",
    "KEYCODE_DPAD_RIGHT",
    "KEYCODE_DPAD_UP",
    "KEYCODE_DPAD_UP_LEFT",
    "KEYCODE_DPAD_UP_RIGHT",
    "KEYCODE_DVR",
    "KEYCODE_E",
    "KEYCODE_EISU",
    "KEYCODE_ENDCALL",
    "KEYCODE_ENTER",
    "KEYCODE_ENVELOPE",
    "KEYCODE_EQUALS",
    "KEYCODE_ESCAPE",
    "KEYCODE_EXPLORER",
    "KEYCODE_F",
    "KEYCODE_F1",
    "KEYCODE_F10",
    "KEYCODE_F11",
    "KEYCODE_F12",
    "KEYCODE_F2",
    "KEYCODE_F3",
    "KEYCODE_F4",
    "KEYCODE_F5",
    "KEYCODE_F6",
    "KEYCODE_F7",
    "KEYCODE_F8",
    "KEYCODE_F9",
    "KEYCODE_FOCUS",
    "KEYCODE_FORWARD",
    "KEYCODE_FORWARD_DEL",
    "KEYCODE_FUNCTION",
    "KEYCODE_G",
    "KEYCODE_GRAVE",
    "KEYCODE_GUIDE",
    "KEYCODE_H",
    "KEYCODE_HEADSETHOOK",
    "KEYCODE_HELP",
    "KEYCODE_HENKAN",
    "KEYCODE_HOME",
    "KEYCODE_I",
    "KEYCODE_INFO",
    "KEYCODE_INSERT",
    "KEYCODE_J",
    "KEYCODE_K",
    "KEYCODE_KANA",
    "KEYCODE_KATAKANA_HIRAGANA",
    "KEYCODE_L",
    "KEYCODE_LANGUAGE_SWITCH",
    "KEYCODE_LAST_CHANNEL",
    "KEYCODE_LEFT_BRACKET",
    "KEYCODE_M",
    "KEYCODE_MANNER_MODE",
    "KEYCODE_MEDIA_AUDIO_TRACK",
    "KEYCODE_MEDIA_CLOSE",
    "KEYCODE_MEDIA_EJECT",
    "KEYCODE_MEDIA_FAST_FORWARD",
    "KEYCODE_MEDIA_NEXT",
    "KEYCODE_MEDIA_PAUSE",
    "KEYCODE_MEDIA_PLAY",
    "KEYCODE_MEDIA_PLAY_PAUSE",
    "KEYCODE_MEDIA_PREVIOUS",
    "KEYCODE_MEDIA_RECORD",
    "KEYCODE_MEDIA_REWIND",
    "KEYCODE_MEDIA_SKIP_BACKWARD",
    "KEYCODE_MEDIA_SKIP_FORWARD",
    "KEYCODE_MEDIA_STEP_BACKWARD",
    "KEYCODE_MEDIA_STEP_FORWARD",
    "KEYCODE_MEDIA_STOP",
    "KEYCODE_MEDIA_TOP_MENU",
    "KEYCODE_MENU",
    "KEYCODE_META_LEFT",
    "KEYCODE_META_RIGHT",
    "KEYCODE_MINUS",
    "KEYCODE_MOVE_END",
    "KEYCODE_MOVE_HOME",
    "KEYCODE_MUHENKAN",
    "KEYCODE_MUSIC",
    "KEYCODE_MUTE",
    "KEYCODE_N",
    "KEYCODE_NAVIGATE_IN",
    "KEYCODE_NAVIGATE_NEXT",
    "KEYCODE_NAVIGATE_OUT",
    "KEYCODE_NAVIGATE_PREVIOUS",
    "KEYCODE_NOTIFICATION",
    "KEYCODE_NUM",
    "KEYCODE_NUM_LOCK",
    "KEYCODE_NUMPAD_0",
    "KEYCODE_NUMPAD_1",
    "KEYCODE_NUMPAD_2",
    "KEYCODE_NUMPAD_3",
    "KEYCODE_NUMPAD_4",
    "KEYCODE_NUMPAD_5",
    "KEYCODE_NUMPAD_6",
    "KEYCODE_NUMPAD_7",
    "KEYCODE_NUMPAD_8",
    "KEYCODE_NUMPAD_9",
    "KEYCODE_NUMPAD_ADD",
    "KEYCODE_NUMPAD_COMMA",
    "KEYCODE_NUMPAD_DIVIDE",
    "KEYCODE_NUMPAD_DOT",
    "KEYCODE_NUMPAD_ENTER",
    "KEYCODE_NUMPAD_EQUALS",
    "KEYCODE_NUMPAD_LEFT_PAREN",
    "KEYCODE_NUMPAD_MULTIPLY",
    "KEYCODE_NUMPAD_RIGHT_PAREN",
    "KEYCODE_NUMPAD_SUBTRACT",
    "KEYCODE_O",
    "KEYCODE_P",
    "KEYCODE_PAGE_DOWN",
    "KEYCODE_PAGE_UP",
    "KEYCODE_PAIRING",
    "KEYCODE_PASTE",
    "KEYCODE_PERIOD",
    "KEYCODE_PICTSYMBOLS",
    "KEYCODE_PLUS",
    "KEYCODE_POUND",
    "KEYCODE_POWER",
    "KEYCODE_PROG_BLUE",
    "KEYCODE_PROG_GREEN",
    "KEYCODE_PROG_RED",
    "KEYCODE_PROG_YELLOW",
    "KEYCODE_Q",
    "KEYCODE_R",
    "KEYCODE_RIGHT_BRACKET",
    "KEYCODE_RO",
    "KEYCODE_S",
    "KEYCODE_SCROLL_LOCK",
    "KEYCODE_SEARCH",
    "KEYCODE_SEMICOLON",
    "KEYCODE_SETTINGS",
    "KEYCODE_SHIFT_LEFT",
    "KEYCODE_SHIFT_RIGHT",
    "KEYCODE_SLASH",
    "KEYCODE_SLEEP",
    "KEYCODE_SOFT_LEFT",
    "KEYCODE_SOFT_RIGHT",
    "KEYCODE_SOFT_SLEEP",
    "KEYCODE_SPACE",
    "KEYCODE_STAR",
    "KEYCODE_STB_INPUT",
    "KEYCODE_STB_POWER",
    "KEYCODE_STEM_1",
    "KEYCODE_STEM_2",
    "KEYCODE_STEM_3",
    "KEYCODE_STEM_PRIMARY",
    "KEYCODE_SWITCH_CHARSET",
    "KEYCODE_SYM",
    "KEYCODE_SYSRQ",
    "KEYCODE_SYSTEM_NAVIGATION_DOWN",
    "KEYCODE_SYSTEM_NAVIGATION_LEFT",
    "KEYCODE_SYSTEM_NAVIGATION_RIGHT",
    "KEYCODE_SYSTEM_NAVIGATION_UP",
    "KEYCODE_T",
    "KEYCODE_TAB",
    "KEYCODE_TV",
    "KEYCODE_TV_ANTENNA_CABLE",
    "KEYCODE_TV_AUDIO_DESCRIPTION",
    "KEYCODE_TV_AUDIO_DESCRIPTION_MIX_DOWN",
    "KEYCODE_TV_AUDIO_DESCRIPTION_MIX_UP",
    "KEYCODE_TV_CONTENTS_MENU",
    "KEYCODE_TV_DATA_SERVICE",
    "KEYCODE_TV_INPUT",
    "KEYCODE_TV_INPUT_COMPONENT_1",
    "KEYCODE_TV_INPUT_COMPONENT_2",
    "KEYCODE_TV_INPUT_COMPOSITE_1",
    "KEYCODE_TV_INPUT_COMPOSITE_2",
    "KEYCODE_TV_INPUT_HDMI_1",
    "KEYCODE_TV_INPUT_HDMI_2",
    "KEYCODE_TV_INPUT_HDMI_3",
    "KEYCODE_TV_INPUT_HDMI_4",
    "KEYCODE_TV_INPUT_VGA_1",
    "KEYCODE_TV_MEDIA_CONTEXT_MENU",
    "KEYCODE_TV_NETWORK",
    "KEYCODE_TV_NUMBER_ENTRY",
    "KEYCODE_TV_POWER",
    "KEYCODE_TV_RADIO_SERVICE",
    "KEYCODE_TV_SATELLITE",
    "KEYCODE_TV_SATELLITE_BS",
    "KEYCODE_TV_SATELLITE_CS",
    "KEYCODE_TV_SATELLITE_SERVICE",
    "KEYCODE_TV_TELETEXT",
    "KEYCODE_TV_TERRESTRIAL_ANALOG",
    "KEYCODE_TV_TERRESTRIAL_DIGITAL",
    "KEYCODE_TV_TIMER_PROGRAMMING",
    "KEYCODE_TV_ZOOM_MODE",
    "KEYCODE_U",
    "KEYCODE_V",
    "KEYCODE_VOICE_ASSIST",
    "KEYCODE_VOLUME_DOWN",
    "KEYCODE_VOLUME_MUTE",
    "KEYCODE_VOLUME_UP",
    "KEYCODE_W",
    "KEYCODE_WAKEUP",
    "KEYCODE_WINDOW",
    "KEYCODE_X",
    "KEYCODE_Y",
    "KEYCODE_YEN",
    "KEYCODE_Z",
    "KEYCODE_ZENKAKU_HANKAKU",
    "KEYCODE_ZOOM_IN",
    "KEYCODE_ZOOM_OUT",
]


# Map a few standard Stb-tester key names to Android keycodes.
# So far we just map the buttons on the Amazon Fire TV & NVidia Shield remote
# controls:
# https://developer.amazon.com/docs/fire-tv/remote-input.html#input-event-reference
_KEYCODE_MAPPINGS = {
    "KEY_BACK": "KEYCODE_BACK",
    "KEY_DOWN": "KEYCODE_DPAD_DOWN",
    "KEY_FASTFORWARD": "KEYCODE_MEDIA_FAST_FORWARD",
    "KEY_HOME": "KEYCODE_HOME",
    "KEY_LEFT": "KEYCODE_DPAD_LEFT",
    "KEY_MENU": "KEYCODE_MENU",
    "KEY_MUTE": "KEYCODE_VOLUME_MUTE",
    "KEY_OK": "KEYCODE_ENTER",
    "KEY_PAUSE": "KEYCODE_MEDIA_PAUSE",
    "KEY_PLAY": "KEYCODE_MEDIA_PLAY",
    "KEY_PLAYPAUSE": "KEYCODE_MEDIA_PLAY_PAUSE",
    "KEY_POWER": "KEYCODE_POWER",
    "KEY_REWIND": "KEYCODE_MEDIA_REWIND",
    "KEY_RIGHT": "KEYCODE_DPAD_RIGHT",
    "KEY_UP": "KEYCODE_DPAD_UP",
    "KEY_VOLUMEDOWN": "KEYCODE_VOLUME_DOWN",
    "KEY_VOLUMEUP": "KEYCODE_VOLUME_UP",
}


def _is_ip_address(address):
    """
    >>> _is_ip_address("")
    False
    >>> _is_ip_address(None)
    False
    >>> _is_ip_address("7278681B045C937CEB770FD31542B16")
    False
    >>> _is_ip_address("192.168.2.11")
    True
    >>> _is_ip_address("192.168.2.11:5555")
    True
    >>> _is_ip_address("192.168.2.11:5555")
    True
    """
    return bool(re.match(r"^\d+\.\d+\.\d+\.\d+(:\d+)?$", str(address)))


def _resize(img, coordinate_system):
    import cv2
    import numpy

    w, h = img.shape[1], img.shape[0]

    if coordinate_system == CoordinateSystem.ADB_NATIVE:
        pass
    elif coordinate_system == CoordinateSystem.ADB_720P:
        # Resize to 720p preserving orientation
        if w > h:
            img = cv2.resize(img, (1280, 720))
        else:
            img = cv2.resize(img, (720, 1280))
    elif coordinate_system == CoordinateSystem.HDMI_720P:
        if w > h:
            # Landscape: The device's screen fills the HDMI frame.
            # (this assumes that the device's aspect ratio is 16:9).
            img = cv2.resize(img, (1280, 720))
        else:
            # Portrait image in a landscape frame, with black letterboxing
            # on either side.
            ratio = float(h) / 720
            w_new = int(w // ratio)
            img = cv2.resize(img, (w_new, 720))
            left = (1280 - w_new) // 2
            img = cv2.copyMakeBorder(img, 0, 0, left, 1280 - w_new - left,
                                     cv2.BORDER_CONSTANT, (0, 0, 0))
    elif coordinate_system == CoordinateSystem.CAMERA_720P:
        # Resize to 720p landscape for compatibility with screenshots from
        # `stbt.get_frame` with the Stb-tester CAMERA.
        if w > h:
            img = cv2.resize(img, (1280, 720))
        else:
            img = numpy.rot90(cv2.resize(img, (720, 1280)))
    else:
        raise NotImplementedError(
            "AdbDevice.get_frame not implemented for %s. "
            "Use coordinate_system=CoordinateSystem.ADB_NATIVE"
            % coordinate_system)

    return img


def _centre_point(r):
    try:
        if all(hasattr(r, name) for name in ("x", "y", "right", "bottom")):
            return (int((r.x + r.right) // 2), int((r.y + r.bottom) // 2))
        elif isinstance(r, tuple) and len(r) == 2:
            return (int(r[0]), int(r[1]))
    except (TypeError, ValueError):
        pass
    raise TypeError("Expected stbt.Region or (x,y) tuple but got %r" % (r,))


_Dimensions = namedtuple("Dimensions", "width height")


def _to_native_coordinates(x, y, coordinate_system, device):
    if coordinate_system == CoordinateSystem.ADB_720P:
        # x & y orientation matches device's orientation
        if device.width > device.height:
            return int(x * device.width / 1280), int(y * device.height / 720)
        else:
            return int(x * device.width / 720), int(y * device.height / 1280)
    elif coordinate_system == CoordinateSystem.HDMI_720P:
        if device.width > device.height:
            # Landscape: Assume the device's native resolution is the same
            # aspect ratio as the HDMI output (i.e. no letterboxing).
            return int(x * device.width / 1280), int(y * device.height / 720)
        else:
            # Portrait image within a landscape screenshot, with black
            # letterboxing on either side. If we assume the device is 16:9
            # aspect ratio and fills the frame vertically, then the device
            # fills pixels 438 to 842 horizontally (404px wide).
            #
            # /-----------------\
            # |######     ######|  ^
            # |######     ######| 720
            # |######     ######|
            # |######     ######|  v
            # \-----------------/
            #        <404>
            #
            if not 438 <= x < 842:
                raise ValueError(
                    "Coordinates %d,%d are outside of the image area that "
                    "corresponds to the device under test" % (x, y))
            x = (x - 438) * 720 / 404
            return int(x * device.width / 720), int(y * device.height / 720)
    elif coordinate_system == CoordinateSystem.CAMERA_720P:
        # x & y coordinates are always from a 720p landscape screenshot.
        if device.width > device.height:
            # Landscape: Device orientation matches x & y.
            return int(x * device.width / 1280), int(y * device.height / 720)
        else:
            # Portrait: The image is rotated 90° anti-clockwise. The device's
            # origin (top-left) corresponds to the bottom-left corner of the
            # screenshot.
            #
            # /--------------------\    /-----\
            # |                    |    |     |
            # |                    |    |     |
            # |                    | => |%    |
            # |                    |    |     |
            # |  %                 |    |     |
            # \--------------------/    ~     ~
            #
            x, y = 720 - y, x
            return int(x * device.width / 720), int(y * device.height / 1280)
    else:
        raise NotImplementedError(
            "AdbDevice: Mapping to native coordinates not implemented for %s"
            % coordinate_system)


def _parse_display_dimensions(dumpsys_output):
    in_display = False
    for line in dumpsys_output.split("\n"):
        if "Display:" in line:
            in_display = True
        if not in_display:
            continue
        m = re.search(r"cur=(\d+)x(\d+)", line)
        if m:
            return _Dimensions(width=int(m.group(1)), height=int(m.group(2)))
    raise AdbError("Didn't find display size in dumpsys output",
                   output=dumpsys_output)



from __future__ import annotations

from dataclasses import dataclass
from functools import lru_cache
from typing import TypeAlias

import cv2
import numpy

from .imgutils import (
    _convert_color, crop, find_file, Image, ImageT, _image_region, load_image,
    pixel_bounding_box, _relative_filename)
from .logging import logger
from .types import Region

try:
    from _stbt.xxhash import Xxhash64
except ImportError:
    Xxhash64 = None


MaskTypes: TypeAlias = "str | numpy.ndarray | Mask | Region | None"


def load_mask(mask: MaskTypes) -> Mask:
    """Used to load a mask from disk, or to create a mask from a `Region`.

    A mask is a black & white image (the same size as the video-frame) that
    specifies which parts of the frame to process: White pixels select the area
    to process, black pixels the area to ignore.

    In most cases you don't need to call ``load_mask`` directly; Stb-tester's
    image-processing functions such as `is_screen_black`, `press_and_wait`, and
    `wait_for_motion` will call ``load_mask`` with their ``mask`` parameter.
    This function is a public API so that you can use it if you are
    implementing your own image-processing functions.

    Note that you can pass a `Region` directly to the ``mask`` parameter of
    stbt functions, and you can create more complex masks by adding,
    subtracting, or inverting Regions (see :doc:`masks`).

    :param str|Region mask: A relative or absolute filename of a mask PNG
      image. If given a relative filename, this uses the algorithm from
      `load_image` to find the file.

      Or, a `Region` that specifies the area to process.

    :returns: A mask as used by `is_screen_black`, `press_and_wait`,
      `wait_for_motion`, and similar image-processing functions.

    Added in v33.
    """
    if mask is None:
        raise TypeError(
            "'mask=None' means an empty region. To analyse the entire "
            "frame use 'mask=Region.ALL' (which is the default)")
    elif isinstance(mask, Mask):
        return mask
    else:
        return Mask(mask)


class Mask:
    """Internal representation of a mask.

    Most users will never need to use this type directly; instead, pass a
    filename or a `Region` to the ``mask`` parameter of APIs like
    `stbt.wait_for_motion`. See :doc:`masks`.
    """
    def __init__(self, m: MaskTypes, *, invert: bool = False) -> None:
        """Private constructor; for public use see `load_mask`."""
        # One (and only one) of these will be set: filename, array, binop,
        # region.
        self._filename = None
        self._array = None
        self._binop = None
        self._region = None
        if isinstance(m, str):
            absolute_filename = find_file(m)
            self._filename = absolute_filename
            self._invert = invert
        elif isinstance(m, numpy.ndarray):
            self._array = load_image(m, color_channels=(1, 3))
            self._invert = invert
        elif isinstance(m, BinOp):
            self._binop = m
            self._invert = invert
        elif isinstance(m, Mask):
            self._filename = m._filename
            self._array = m._array
            self._binop = m._binop
            self._region = m._region
            self._invert = m._invert
            if invert:
                self._invert = not self._invert
        elif isinstance(m, Region):
            self._region = m
            self._invert = invert
        elif m is None:  # Region.intersect can return None for "no region"
            self._region = Region.ALL
            self._invert = not invert
        else:
            raise TypeError("Expected filename, Image, Mask, or Region. "
                            f"Got {m!r}")

    @staticmethod
    def from_alpha_channel(image: ImageT) -> Mask:
        """Create a mask from the alpha channel of an image.

        :type image: string or `numpy.ndarray`
        :param image:
          An image with an alpha (transparency) channel. This can be the
          filename of a png file on disk, or an image previously loaded with
          `stbt.load_image`.

          Filenames should be relative paths. See `stbt.load_image` for the
          path lookup algorithm.
        """
        image = load_image(image)
        if image.shape[2] == 4:
            return Mask(image[:, :, 3])
        else:
            return Mask(_image_region(image))

    def __eq__(self, o):
        if isinstance(o, Region):
            return self.__eq__(Mask(o))
        if not isinstance(o, Mask):
            return False
        if self._array is not None:
            return numpy.array_equal(self._array, o._array)
        else:
            return ((self._filename, self._binop, self._region, self._invert) ==
                    (o._filename, o._binop, o._region, o._invert))

    def __hash__(self):
        if self._region is not None and not self._invert:
            return hash(self._region)
        elif self._array is not None:
            if Xxhash64:
                h = Xxhash64()
                h.update(numpy.ascontiguousarray(self._array).data)
                digest = h.hexdigest()
            else:
                digest = hash(self._array.data.tobytes())
            return hash((self._array.shape, digest, self._invert))
        else:
            return hash((self._filename, self._binop, self._region,
                         self._invert))

    def to_array(self, region: Region, color_channels: int = 1) \
            -> tuple[numpy.ndarray | None, Region]:
        """Materialize the mask to a numpy array of the specified size.

        Most users will never need to call this method; it's for people who
        are implementing their own image-processing algorithms.

        :param stbt.Region region: A Region matching the size of the frame that
          you are processing.

        :param int color_channels: The number of channels required (1 or 3),
          according to your image-processing algorithm's needs. All channels
          will be identical — for example with 3 channels, pixels will be
          either [0, 0, 0] or [255, 255, 255].

        :rtype: tuple[numpy.ndarray | None, Region]
        :returns:
          A tuple of:

          * An image (numpy array), where masked-in pixels are white (255) and
            masked-out pixels are black (0). The array is the same size as
            the region in the second member of this tuple.
          * A bounding box (`stbt.Region`) around the masked-in area. If most
            of the frame is masked out, limiting your image-processing
            operations to this region will be faster.

          If the mask is just a Region, the first member of the tuple (the
          image) will be `None` because the bounding-box is sufficient.
        """
        if region.x != 0 or region.y != 0:
            raise ValueError(
                f"{region} must be a full-frame region starting at x=0, y=0")
        if color_channels not in (1, 3):
            raise ValueError(f"Invalid {color_channels=} (expected 1 or 3)")
        return _to_array_and_bounding_box_cached(self, region, color_channels)

    def __repr__(self):
        # In-order traversal, removing unnecessary parentheses.
        prefix = "~" if self._invert else ""
        if self._filename is not None:
            return f"{prefix}Mask({_relative_filename(self._filename)!r})"
        elif self._array is not None:
            if isinstance(self._array, Image) and self._array.relative_filename:
                return f"{prefix}Mask({self._array.relative_filename!r})"
            else:
                return f"{prefix}Mask(<Image>)"
        elif self._binop is not None:
            left_repr = repr(self._binop.left)
            right_repr = repr(self._binop.right)
            if "-" in right_repr and right_repr[0] not in ("~", "("):
                right_repr = f"({right_repr})"
            if prefix:
                open_paren, close_paren = "(", ")"
            else:
                open_paren, close_paren = "", ""
            return (f"{prefix}{open_paren}{left_repr} {self._binop.op} "
                    f"{right_repr}{close_paren}")
        elif self._region is not None:
            return f"{prefix}{self._region!r}"
        else:
            assert False, "Internal logic error"

    def __add__(self, other: MaskTypes) -> Mask:
        if isinstance(other, (Region, Mask, type(None))):
            return Mask(BinOp("+", self, Mask(other)))
        else:
            return NotImplemented

    def __radd__(self, other: MaskTypes) -> Mask:
        if isinstance(other, (Region, Mask, type(None))):
            return Mask(other).__add__(self)
        else:
            return NotImplemented

    def __sub__(self, other: MaskTypes) -> Mask:
        if isinstance(other, (Region, Mask, type(None))):
            return Mask(BinOp("-", self, Mask(other)))
        else:
            return NotImplemented

    def __rsub__(self, other: MaskTypes) -> Mask:
        if isinstance(other, (Region, Mask, type(None))):
            return Mask(other).__sub__(self)
        else:
            return NotImplemented

    def __invert__(self) -> Mask:
        return Mask(self, invert=True)


@dataclass(frozen=True)
class BinOp:
    op: str  # "+" or "-"
    left: Mask
    right: Mask


@lru_cache(maxsize=10)
def _to_array_and_bounding_box_cached(
        mask: Mask,
        region: Region,
        color_channels: int) -> tuple[numpy.ndarray | None, Region]:

    if mask._region is not None and not mask._invert:
        array = None
    else:
        array = _to_array(mask, region)

    if mask._region is Region.ALL:
        if mask._invert:
            bounding_box = None
        else:
            bounding_box = region
    elif mask._region is not None and not mask._invert:
        bounding_box = Region.intersect(region, mask._region)
    else:
        bounding_box = pixel_bounding_box(array)
        if bounding_box is None:
            logger.warning("%r is an empty Region", mask)

    if bounding_box is None:
        raise ValueError("%r doesn't overlap with the frame's %r"
                         % (mask, region))

    if array is not None:
        array = crop(array, bounding_box)
        nonzeros = numpy.count_nonzero(array)
        if nonzeros == array.size:
            # Every pixel is masked in so the pixel-mask is redundant.
            array = None
    if array is not None:
        if color_channels == 3:
            array = cv2.cvtColor(array, cv2.COLOR_GRAY2BGR)
        array.flags.writeable = False

    return array, bounding_box


def _to_array(mask: Mask, region: Region) -> numpy.ndarray:
    array: numpy.ndarray
    shape = (region.height, region.width, 1)
    if mask._filename is not None:
        array = load_image(mask._filename, color_channels=(1,))
        if array.shape != shape:
            raise ValueError(f"{mask}: shape {array.shape} doesn't match "
                             f"required shape {shape}")
    elif mask._array is not None:
        array = mask._array
        array = _convert_color(array, color_channels=(1,),
                               absolute_filename=array.absolute_filename)
        if array.shape != shape:
            raise ValueError(f"{mask}: shape {array.shape} doesn't match "
                             f"required shape {shape}")
    elif mask._binop is not None:
        n = mask._binop
        if n.op == "+":
            array = _to_array(n.left, region) | _to_array(n.right, region)
        elif n.op == "-":
            array = _to_array(n.left, region) & ~_to_array(n.right, region)  # pylint:disable=invalid-unary-operand-type
        else:
            assert False, f"Unreachable: Unknown op {n.op}"
    elif mask._region is not None:
        array = numpy.full(shape, 0, dtype=numpy.uint8)
        r = Region.intersect(mask._region, region)
        if r:
            array[r.y:r.bottom, r.x:r.right] = 255
    else:
        assert False, "Internal logic error"

    if mask._invert:
        array = ~array  # pylint:disable=invalid-unary-operand-type

    return array
